"""
Use `list_devices` to discover input devices connected to your machine.
Each device is an instance of either `Mouse`, `Keyboard` or `HID` (experimental).

Use `hook_raw_input_for_window` to receive raw input events.
Each event also references the `RawInputDevice` that triggered it.
"""
from ._api import hook_raw_input_for_window, RawInputDevice, Mouse, Keyboard, HID, RawInputEvent, list_devices


__all__ = [key for key in globals().keys() if not key.startswith('_')]



import ctypes

# Define constants and structures for Windows API calls
HIDP_STATUS_SUCCESS = 0x110000
MAX_STRING_LENGTH = 126
DIGCF_PRESENT = 0x00000002
DIGCF_DEVICEINTERFACE = 0x00000010



class GUID(ctypes.Structure):
    _fields_ = [
        ("Data1", ctypes.c_ulong),
        ("Data2", ctypes.c_ushort),
        ("Data3", ctypes.c_ushort),
        ("Data4", ctypes.c_ubyte * 8)
    ]


class SP_DEVICE_INTERFACE_DATA(ctypes.Structure):
    _fields_ = [
        ("cbSize", ctypes.c_ulong),
        ("InterfaceClassGuid", GUID),
        ("Flags", ctypes.c_ulong),
        ("Reserved", ctypes.POINTER(ctypes.c_ulong))
    ]


class SP_DEVINFO_DATA(ctypes.Structure):
    _fields_ = [
        ("cbSize", ctypes.c_ulong),
        ("ClassGuid", GUID),
        ("DevInst", ctypes.c_ulong),
        ("Reserved", ctypes.POINTER(ctypes.c_ulong))
    ]


class SP_DEVICE_INTERFACE_DETAIL_DATA(ctypes.Structure):
    _fields_ = [
        ("cbSize", ctypes.c_ulong),
        ("DevicePath", ctypes.c_char * 1)
    ]


input_id = "\\\\?\\HID#VID_046D&PID_C336&MI_00#8&178e0225&0&0000#{884b96c3-56ef-11d1-bc8c-00a0c91405dd}"

# Load the required DLLs
setupapi = ctypes.windll.LoadLibrary("setupapi")
hid = ctypes.windll.LoadLibrary("hid")

# Get the HID GUID
hid_guid = GUID()
hid.HidD_GetHidGuid(ctypes.byref(hid_guid))

# Get a handle to a device information set for all HID devices
# devinfo = setupapi.SetupDiGetClassDevsA(ctypes.byref(hid_guid), None, None, DIGCF_PRESENT | DIGCF_DEVICEINTERFACE)
devinfo = setupapi.SetupDiGetClassDevsA(ctypes.byref(hid_guid), None, None, DIGCF_PRESENT | DIGCF_DEVICEINTERFACE)

# Enumerate the device interfaces in the device information set
index = 0
device_found = False
device_interface_data = SP_DEVICE_INTERFACE_DATA()
device_interface_data.cbSize = ctypes.sizeof(device_interface_data)
while setupapi.SetupDiEnumDeviceInterfaces(devinfo, None, ctypes.byref(hid_guid), index, ctypes.byref(device_interface_data)):
    index += 1
    print("Device")

    # Get the device path for this device interface
    detail_data_size = ctypes.c_ulong()
    setupapi.SetupDiGetDeviceInterfaceDetailA(devinfo, ctypes.byref(device_interface_data), None, 0, ctypes.byref(detail_data_size), None)
    detail_data_buffer = ctypes.create_string_buffer(detail_data_size.value)
    detail_data = ctypes.cast(detail_data_buffer, ctypes.POINTER(SP_DEVICE_INTERFACE_DETAIL_DATA)).contents
    detail_data.cbSize = 5  # sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA)
    setupapi.SetupDiGetDeviceInterfaceDetailA(devinfo, ctypes.byref(device_interface_data), ctypes.byref(detail_data), detail_data_size, None, None)

    # Get the device id for this device interface
    device_id_buffer = ctypes.create_string_buffer(MAX_STRING_LENGTH)
    hid.HidD_GetHidGuid(ctypes.byref(device_id_buffer))

    # Check if the device id matches the input id
    device_id = device_id_buffer.value.decode("utf-8")
    if device_id == input_id:
        print(device_id)
        device_found = True
        break
else:
    raise ctypes.WinError(ctypes.GetLastError())

# If the device is found, get its manufacturer and product name
if device_found:
    # Open the device with read access
    device_handle = setupapi.CreateFileA(detail_data.DevicePath, 0x80000000, 0, None, 3, 0, None)

    # Get the manufacturer name
    manufacturer_buffer = ctypes.create_string_buffer(MAX_STRING_LENGTH)
    hid.HidD_GetManufacturerString(device_handle, manufacturer_buffer, MAX_STRING_LENGTH)

    # Get the product name
    product_buffer = ctypes.create_string_buffer(MAX_STRING_LENGTH)
    hid.HidD_GetProductString(device_handle, product_buffer, MAX_STRING_LENGTH)

    # Close the device handle
    setupapi.CloseHandle(device_handle)

    # Print the manufacturer and product name
    manufacturer_name = manufacturer_buffer.value.decode("utf-16")
    product_name = product_buffer.value.decode("utf-16")

    print(f"Manufacturer: {manufacturer_name}")
    print(f"Product: {product_name}")
else:
    print(f"No device with id {input_id} found")

# Destroy the device information set
setupapi.SetupDiDestroyDeviceInfoList(devinfo)


import time
from dataclasses import dataclass
from typing import Optional, Sequence

from ._win_raw_input import *
from .usb_ids import parse_vid_pid, lookup_product


@dataclass
class RawInputDevice:
    """
    Base class for input devices.

    There are three types of raw input devices:

    * `Mouse`
    * `Keyboard`
    * `HID`

    Some devices cannot be identified, e.g. events caused by software.
    Then the devices stores an invalid handle and has `path=None`.
    """

    handle: int
    """Windows handle for the device. This is valid only in the current session. For a unique identifier, use `path`."""
    path: Optional[str]
    """Unique device identifier. This string typically includes vendor and product id."""
    vendor_id: int
    """USB vendor id"""
    vendor_name: Optional[str]
    """Vendor name or `None` if unknown."""
    product_id: int
    """USB product id"""
    product_name: str
    """Product name or `None` if unknown."""

    def is_connected(self):
        """
        Checks whether the handle associated with this device is still valid.
        If not, the device has most likely been disconnected.
        """
        return is_connected(self.handle)


@dataclass
class Keyboard(RawInputDevice):
    """
    Keyboards send events when a key is pressed, released or while a key is being held.
    All events have either `event_type='down'` for key pressed/held or `event_type='up'` for key released.
    The `RawInputEvent.code` denotes the scan code and `RawInputEvent.name` the human-readable name of the key.
    If a keypad key is pressed, the `RawInputEvent.device_type` is set to `'keypad'`.

    This class inherits the method `RawInputDevice.is_connected()`.
    """
    keyboard_type: str
    subtype: int
    scan_code_mode: int
    num_function_keys: int
    num_indicators: int
    num_keys: int

    def __hash__(self):
        return self.handle or 0

    def __eq__(self, other):
        return isinstance(other, Keyboard) and other.handle == self.handle


@dataclass
class Mouse(RawInputDevice):
    """
    Mice can send the following event types:

    * `'down'` when a button is pressed
    * `'up'` when a button is released
    * `'move'` when the mouse is moved
    * `'wheel-up'` when the mouse is turned upwards
    * `'wheel-down'` when the mouse is turned down

    In case of `move` events, `RawInputEvent.delta_x` and `RawInputEvent.delta_y` hold the amount by which the mouse moved.
    For button events, `code` denotes the button index and `name` stores the human-readable button name.

    This class inherits the method `RawInputDevice.is_connected()`.
    """
    mouse_type: str
    """HID mouse, HID wheel mouse, Mouse with horizontal wheel, unknown"""
    num_buttons: int
    sample_rate: Optional[int]
    """data points per second (if available)"""
    has_horizontal_wheel: bool

    def __hash__(self):
        return self.handle or 0

    def __eq__(self, other):
        return isinstance(other, Mouse) and other.handle == self.handle


@dataclass
class HID(RawInputDevice):
    """
    HID devices that are neither classified as keyboard nor mouse.
    These devices send raw data as bytes, stored in `RawInputEvent.data`.
    All input events are classified as `event_type='data'`.

    This class inherits the method `RawInputDevice.is_connected()`.
    """
    version_number: int
    usage_page: int
    usage_page_name: str
    usage: int
    usage_name: str

    def __hash__(self):
        return self.handle or 0

    def __eq__(self, other):
        return isinstance(other, HID) and other.handle == self.handle


CACHED_DEVICES = {}  # handle -> RawInputDevice


def list_devices() -> Sequence[RawInputDevice]:
    """
    Retrieves a list of all currently registered raw input devices using `GetRawInputDeviceList`.
    For each device, this function also looks up its unique name and additional information.

    All discovered devices are cached by their `handle` which is unique during this session.

    Returns:
        Devices as sequence of `RawInputDevice`
    """
    devices = [get_device(dev.dwType, dev.hDevice) for dev in get_raw_input_device_list()]
    CACHED_DEVICES.clear()
    CACHED_DEVICES.update({d.handle: d for d in devices})
    return devices


def get_device(dw_type: int, handle: int) -> RawInputDevice:
    if handle in CACHED_DEVICES:
        return CACHED_DEVICES[handle]
    path = get_device_path(handle)
    info = get_device_info(handle)
    if dw_type == 0:
        mouse_type = MOUSE_TYPES.get(info.u.mouse.dwId, 'unknown')
        num_buttons = info.u.mouse.dwNumberOfButtons
        sample_rate = info.u.mouse.dwSampleRate or None
        has_horizontal_wheel = bool(info.u.mouse.fHasHorizontalWheel)
        vendor_id, product_id = parse_vid_pid(path)
        vendor_name, product_name = lookup_product(vendor_id, product_id)
        device = Mouse(handle, path, vendor_id, vendor_name, product_id, product_name, mouse_type, num_buttons, sample_rate, has_horizontal_wheel)
    elif dw_type == 1:
        kb_type = KEYBOARD_TYPES.get(info.u.keyboard.dwType, 'unknown')
        subtype = info.u.keyboard.dwSubType
        scan_code_mode = info.u.keyboard.dwKeyboardMode
        num_function_keys = info.u.keyboard.dwNumberOfFunctionKeys
        num_indicators = info.u.keyboard.dwNumberOfIndicators
        num_keys = info.u.keyboard.dwNumberOfKeysTotal
        vendor_id, product_id = parse_vid_pid(path)
        vendor_name, product_name = lookup_product(vendor_id, product_id)
        device = Keyboard(handle, path, vendor_id, vendor_name, product_id, product_name, kb_type, subtype, scan_code_mode, num_function_keys, num_indicators, num_keys)
    elif dw_type == 2:
        vendor_id = info.u.hid.dwVendorId
        product_id = info.u.hid.dwProductId
        version_number = info.u.hid.dwVersionNumber
        usage_page = info.u.hid.usUsagePage
        usage_page_name, page_entries = USAGE_PAGE_NAMES.get(usage_page, ('unknown', {}))
        usage = info.u.hid.usUsage
        usage_name = page_entries.get(usage, 'unknown')
        vendor_name, product_name = lookup_product(vendor_id, product_id)
        device = HID(handle, path, vendor_id, vendor_name, product_id, product_name, version_number, usage_page, usage_page_name, usage, usage_name)
    else:
        raise NotImplementedError(f"Unknown device type: {dw_type}")
    CACHED_DEVICES[handle] = device
    return device


@dataclass
class RawInputEvent:
    """
    These events are sent to the callback registered via `hook_raw_input_for_window()`.
    """
    event_type: str
    """'up'/'down' for buttons, 'move' for mouse, 'data' for HID"""
    code: int
    """button index or scan code, -1 for HID"""
    name: Optional[str]
    """mouse button or key name"""
    device_type: str
    """'keyboard', 'keypad', 'mouse', 'controller'"""
    device: RawInputDevice
    """Source `RawInputDevice`"""
    delta_x: Optional[int]
    """Horizontal mouse movement"""
    delta_y: Optional[int]
    """Vertical mouse movement"""
    hwnd: int
    """Window handle of the window this event was sent to"""
    time: float
    """Time of the event measured vis `time.perf_counter()`"""
    data: Optional[bytes]
    """Raw data sent by HID devices"""
    raw: RAWINPUT
    """Raw event data sent from Windows"""


def hook_raw_input_for_window(hwnd,
                              callback: Callable[[RawInputEvent], None],
                              device_types=('Pointer', 'Mouse', 'Joystick', 'Game Pad', 'Keyboard', 'Keypad', 'Multi-axis Controller')):
    """
    Listen to raw input events sent to a window by the operating system (Windows).

    This is implemented by overriding the window's procedure function which is the only function that receives raw input events.
    The new procedure calls the callback before executing the regular window procedure.

    The HWND can be retrieved from a TK instance using `window.winfo_id()` and similar functions exist for other libraries.

    Args:
        hwnd: Window handle (HWND)
            This can be obtained from various libraries that open windows, such as

            * `window.winfo_id()` in `tkinter`
            * `pygame.display.get_wm_info()['window']` in PyGame
            * Most functions in win32gui
            * `fig.canvas.manager.window.winId()` in matplotlib using PyQt5

        callback: Function
        device_types:
           Types of devices to listen for.
           This defines the usage page and value.
           Supported device types are `'Pointer'`, `'Mouse'`, `'Joystick'`, `'Game Pad'`, `'Keyboard'`, `'Keypad'`, `'Multi-axis Controller'`.
    """
    def process_message(hwnd, msg, wParam, lParam):  # Called by Windows to handle window events.
        if msg == WM_INPUT:  # raw input. other events don't reference the device
            event_time = time.perf_counter()
            dwSize = c_uint()
            if user32.GetRawInputData(lParam, RID_INPUT, NULL, byref(dwSize), sizeof(RAWINPUTHEADER)):
                raise ctypes.WinError(GetLastError())
            raw = RAWINPUT()
            assert user32.GetRawInputData(lParam, RID_INPUT, byref(raw), byref(dwSize), sizeof(RAWINPUTHEADER)) == dwSize.value
            device = get_device(raw.header.dwType, raw.header.hDevice)
            if raw.header.dwType == 1:  # Keyboard
                if raw.data.mouse.usFlags in MOVE_MODES and raw.data.mouse._s2.usButtonFlags == 0:  # actually caused by mouse
                    device = get_device(0, raw.header.hDevice)
                assert dwSize.value == 40
                message = raw.data.keyboard.message
                vk_code = raw.data.keyboard.vk_code
                scan_code = raw.data.keyboard.scan_code
                # flags = raw.data.keyboard.flags
                if vk_code in VIRTUAL_KEYBOARD:
                    key_name, is_keypad = VIRTUAL_KEYBOARD[vk_code]
                else:
                    key_name, is_keypad = KEY_NAMES_LOWER[scan_code], False
                evt_type = KEY_EVENT_TYPE[message]
                event = RawInputEvent(evt_type, scan_code, key_name, 'keypad' if is_keypad else 'keyboard', device, None, None, hwnd, event_time, None, raw)
            elif raw.header.dwType == 0:  # Mouse
                assert dwSize.value == 48
                button = raw.data.mouse.ulButtons
                if button == 0:
                    mode = raw.data.mouse.usFlags
                    event = RawInputEvent('move', mode, MOVE_MODES.get(mode, 'unknown'), 'mouse', device, raw.data.mouse.lLastX, raw.data.mouse.lLastY, hwnd, event_time, None, raw)
                else:
                    evt_type, button_id, button_name = MOUSE_BUTTONS[button]
                    event = RawInputEvent(evt_type, button_id, button_name, 'mouse', device, None, None, hwnd, event_time, None, raw)
            elif raw.header.dwType == 2:  # Controller
                size_hid = raw.data.hid.dwSizeHid  # The size, in bytes, of each HID input in bRawData.
                count = raw.data.hid.dwCount  # The number of HID inputs in bRawData.
                raw_data = raw.data.hid.bRawData  # The raw input data, as an array of bytes.
                raw_data_ptr = ctypes.cast(raw_data, POINTER(BYTE * (size_hid * count)))
                raw_data_bytes = ctypes.string_at(raw_data_ptr, sizeof(raw_data_ptr.contents))
                event = RawInputEvent('data', -1, f'{count} x {size_hid} bytes', 'hid', device, None, None, hwnd, event_time, raw_data_bytes, raw)
            else:
                raise NotImplementedError
            callback(event)
    if hwnd:
        set_window_procedure(hwnd, process_message, call_original=True)
    else:
        hwnd = invisible_window(process_message)
    for device_type in device_types:
        enable_raw_input_for_window(hwnd, device_type)



"""
This is the Windows backend for keyboard events, and is implemented by invoking the Win32 API through the ctypes module.
This is error-prone and can introduce very unpythonic failure modes, such as segfaults and low level memory leaks.
But it is also dependency-free, very performant well documented on Microsoft's website and scattered examples.
"""
import ctypes
from ctypes import c_short, c_uint8, c_int, c_uint, c_long, Structure, CFUNCTYPE, POINTER, WINFUNCTYPE, byref, sizeof, Union, c_ushort
from ctypes.wintypes import WORD, DWORD, BOOL, HHOOK, MSG, LPWSTR, WCHAR, WPARAM, LPARAM, LONG, USHORT, HWND, UINT, HANDLE, LPCWSTR, ULONG, BYTE, HMENU, HINSTANCE, LPVOID, INT
from typing import Callable, Optional

user32 = ctypes.windll.user32
kernel32 = ctypes.windll.kernel32

LPMSG = POINTER(MSG)
ULONG_PTR = POINTER(DWORD)


class KBDLLHOOKSTRUCT(Structure):  # https://learn.microsoft.com/en-us/windows/win32/api/winuser/ns-winuser-kbdllhookstruct
    _fields_ = [("vk_code", DWORD),
                ("scan_code", DWORD),
                ("flags", DWORD),
                ("time", c_int),
                ("dwExtraInfo", ULONG_PTR)]


# Included for completeness.
class MOUSEINPUT(Structure):
    _fields_ = (('dx', LONG),
                ('dy', LONG),
                ('mouseData', DWORD),
                ('dwFlags', DWORD),
                ('time', DWORD),
                ('dwExtraInfo', ULONG_PTR))


class KEYBDINPUT(Structure):
    _fields_ = (('vk_code', WORD),
                ('scan_code', WORD),
                ('flags', DWORD),
                ('time', DWORD),
                ('dwExtraInfo', ULONG_PTR))


class HARDWAREINPUT(Structure):
    _fields_ = (('uMsg', DWORD),
                ('wParamL', WORD),
                ('wParamH', WORD))


class _INPUTunion(Union):
    _fields_ = (('mi', MOUSEINPUT),
                ('ki', KEYBDINPUT),
                ('hi', HARDWAREINPUT))


class INPUT(Structure):
    _fields_ = (('type', DWORD),
                ('union', _INPUTunion))


class RAWINPUTDEVICE(Structure):
    _fields_ = [("us_usage_page", USHORT),
                ("us_usage", USHORT),
                ("dw_flags", DWORD),
                ("hwnd_target", HWND)]


class RAWINPUTDEVICELIST(ctypes.Structure):  # https://learn.microsoft.com/en-us/windows/win32/api/winuser/ns-winuser-rawinputdevicelist
    _fields_ = [
        ("hDevice", HANDLE),
        ("dwType", DWORD),
    ]


PRAWINPUTDEVICELIST = POINTER(RAWINPUTDEVICELIST)

WNDPROCTYPE = WINFUNCTYPE(c_int, HWND, UINT, WPARAM, LPARAM)
CS_HREDRAW = 2
CS_VREDRAW = 1
CW_USEDEFAULT = 0x80000000
WHITE_BRUSH = 0

RI_KEY_BREAK = 0x01
RI_KEY_MAKE = 0x00

RIM_TYPES = {0: 'mouse', 1: 'keyboard', 2: 'hid'}

RID_INPUT = 0x10000003

RIDEV_INPUTSINK = 0x00000100


class WNDCLASSEX(Structure):  # https://learn.microsoft.com/en-us/windows/win32/api/winuser/ns-winuser-wndclassexa
    _fields_ = [("cbSize", c_uint),
                ("style", c_uint),
                ("lpfnWndProc", WNDPROCTYPE),
                ("cbClsExtra", c_int),
                ("cbWndExtra", c_int),
                ("hInstance", HANDLE),
                ("hIcon", HANDLE),
                ("hCursor", HANDLE),
                ("hBrush", HANDLE),
                ("lpszMenuName", LPCWSTR),
                ("lpszClassName", LPCWSTR),
                ("hIconSm", HANDLE)]


class RAWINPUTHEADER(Structure):
    _fields_ = [
        ("dwType", DWORD),
        ("dwSize", DWORD),
        ("hDevice", HANDLE),
        ("wParam", WPARAM),
    ]


class RAWMOUSE(Structure):  # https://learn.microsoft.com/en-us/windows/win32/api/winuser/ns-winuser-rawmouse
    class _U1(Union):
        class _S2(Structure):
            _fields_ = [
                ("usButtonFlags", c_ushort),
                ("usButtonData", c_ushort),
            ]

        _fields_ = [
            ("ulButtons", ULONG),
            ("_s2", _S2),
        ]

    _fields_ = [
        ("usFlags", c_ushort),
        ("_u1", _U1),
        ("ulRawButtons", ULONG),
        ("lLastX", LONG),
        ("lLastY", LONG),
        ("ulExtraInformation", ULONG),
    ]
    _anonymous_ = ("_u1",)


class RAWKEYBOARD(Structure):
    _fields_ = [
        ("scan_code", c_ushort),
        ("flags", c_ushort),
        ("reserved", c_ushort),
        ("vk_code", c_ushort),
        ("message", UINT),
        ("dwExtraInfo", ULONG),
    ]


class RAWHID(Structure):  # https://learn.microsoft.com/en-us/windows/win32/api/winuser/ns-winuser-rawhid
    _fields_ = [
        ("dwSizeHid", DWORD),
        ("dwCount", DWORD),
        ("bRawData", BYTE * 1),
    ]


class _RAWINPUTUnion(Union):
    _fields_ = [
        ("mouse", RAWMOUSE),
        ("keyboard", RAWKEYBOARD),
        ("hid", RAWHID),
    ]


class RAWINPUT(Structure):
    _fields_ = [
        ("header", RAWINPUTHEADER),
        ("data", _RAWINPUTUnion),
    ]


class RID_DEVICE_INFO_MOUSE(ctypes.Structure):
    _fields_ = [
        ("dwId", DWORD),
        ("dwNumberOfButtons", DWORD),
        ("dwSampleRate", DWORD),
        ("fHasHorizontalWheel", ctypes.c_int),
    ]


class RID_DEVICE_INFO_KEYBOARD(ctypes.Structure):
    _fields_ = [
        ("dwType", DWORD),
        ("dwSubType", DWORD),
        ("dwKeyboardMode", DWORD),
        ("dwNumberOfFunctionKeys", DWORD),
        ("dwNumberOfIndicators", DWORD),
        ("dwNumberOfKeysTotal", DWORD),
    ]


class RID_DEVICE_INFO_HID(ctypes.Structure):
    _fields_ = [
        ("dwVendorId", DWORD),
        ("dwProductId", DWORD),
        ("dwVersionNumber", DWORD),
        ("usUsagePage", ctypes.c_ushort),
        ("usUsage", ctypes.c_ushort),
    ]


class RID_DEVICE_INFO_UNION(ctypes.Union):
    _fields_ = [
        ("mouse", RID_DEVICE_INFO_MOUSE),
        ("keyboard", RID_DEVICE_INFO_KEYBOARD),
        ("hid", RID_DEVICE_INFO_HID),
    ]


class RID_DEVICE_INFO(ctypes.Structure):
    _fields_ = [
        ("cbSize", DWORD),
        ("dwType", DWORD),
        ("u", RID_DEVICE_INFO_UNION),
    ]


RegisterRawInputDevices = user32.RegisterRawInputDevices
RegisterRawInputDevices.argtypes = [POINTER(RAWINPUTDEVICE), UINT, UINT]
RegisterRawInputDevices.restype = UINT

CreateWindowExW = user32.CreateWindowExW
CreateWindowExW.argtypes = [DWORD, LPCWSTR, LPCWSTR, DWORD, INT, INT, INT, INT, HWND, HMENU, HINSTANCE, LPVOID]
CreateWindowExW.restype = HWND

LowLevelKeyboardProc = CFUNCTYPE(c_int, WPARAM, LPARAM, POINTER(KBDLLHOOKSTRUCT))

SetWindowsHookEx = user32.SetWindowsHookExA  # https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-setwindowshookexa
# SetWindowsHookEx.argtypes = [c_int, LowLevelKeyboardProc, c_int, c_int]
SetWindowsHookEx.restype = HHOOK

CallNextHookEx = user32.CallNextHookEx
# CallNextHookEx.argtypes = [c_int , c_int, c_int, POINTER(KBDLLHOOKSTRUCT)]
CallNextHookEx.restype = c_int

UnhookWindowsHookEx = user32.UnhookWindowsHookEx
UnhookWindowsHookEx.argtypes = [HHOOK]
UnhookWindowsHookEx.restype = BOOL

GetMessage = user32.GetMessageW
GetMessage.argtypes = [LPMSG, c_int, c_int, c_int]
GetMessage.restype = BOOL

GetLastError = kernel32.GetLastError
GetLastError.argtypes = []
GetLastError.restype = DWORD

TranslateMessage = user32.TranslateMessage
TranslateMessage.argtypes = [LPMSG]
TranslateMessage.restype = BOOL

DispatchMessage = user32.DispatchMessageA
DispatchMessage.argtypes = [LPMSG]

keyboard_state_type = c_uint8 * 256

GetKeyboardState = user32.GetKeyboardState
GetKeyboardState.argtypes = [keyboard_state_type]
GetKeyboardState.restype = BOOL

GetKeyNameText = user32.GetKeyNameTextW
GetKeyNameText.argtypes = [c_long, LPWSTR, c_int]
GetKeyNameText.restype = c_int

MapVirtualKey = user32.MapVirtualKeyW
MapVirtualKey.argtypes = [c_uint, c_uint]
MapVirtualKey.restype = c_uint

ToUnicode = user32.ToUnicode
ToUnicode.argtypes = [c_uint, c_uint, keyboard_state_type, LPWSTR, c_int, c_uint]
ToUnicode.restype = c_int

MAPVK_VK_TO_VSC = 0
MAPVK_VSC_TO_VK = 1

VkKeyScan = user32.VkKeyScanW
VkKeyScan.argtypes = [WCHAR]
VkKeyScan.restype = c_short

RIDI_DEVICENAME = 0x20000007
RIDI_DEVICEINFO = 0x2000000b
RIDI_PREPARSEDDATA = 0x20000005

GetRawInputDeviceList = ctypes.windll.user32.GetRawInputDeviceList  # https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-getrawinputdevicelist
GetRawInputDeviceList.argtypes = [PRAWINPUTDEVICELIST, ctypes.POINTER(ctypes.c_uint), ctypes.c_uint]
GetRawInputDeviceList.restype = ctypes.c_uint

GetRawInputDeviceInfoW = ctypes.windll.user32.GetRawInputDeviceInfoW  # https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-getrawinputdeviceinfoa
GetRawInputDeviceInfoW.argtypes = [HANDLE, DWORD, ctypes.c_void_p, ctypes.POINTER(ctypes.c_uint)]
GetRawInputDeviceInfoW.restype = ctypes.c_uint

CallWindowProc = user32.CallWindowProcA
CallWindowProc.argtypes = [HANDLE, HWND, UINT, WPARAM, LPARAM]

NULL = c_int(0)

WM_INPUT = 0x00FF

WM_KEYDOWN = 0x0100
WM_KEYUP = 0x0101
WM_SYSKEYDOWN = 0x104  # Used for ALT key
WM_SYSKEYUP = 0x105


# List taken from the official documentation, but stripped of the OEM-specific keys. Keys are virtual key codes, values are pairs (name, is_keypad).
VIRTUAL_KEYBOARD = {
    0x03: ('control-break processing', False),
    0x08: ('backspace', False),
    0x09: ('tab', False),
    0x0c: ('clear', False),
    0x0d: ('enter', False),
    0x10: ('shift', False),
    0x11: ('ctrl', False),
    0x12: ('alt', False),
    0x13: ('pause', False),
    0x14: ('caps lock', False),
    0x15: ('ime kana mode', False),
    0x15: ('ime hanguel mode', False),
    0x15: ('ime hangul mode', False),
    0x17: ('ime junja mode', False),
    0x18: ('ime final mode', False),
    0x19: ('ime hanja mode', False),
    0x19: ('ime kanji mode', False),
    0x1b: ('esc', False),
    0x1c: ('ime convert', False),
    0x1d: ('ime nonconvert', False),
    0x1e: ('ime accept', False),
    0x1f: ('ime mode change request', False),
    0x20: ('spacebar', False),
    0x21: ('page up', False),
    0x22: ('page down', False),
    0x23: ('end', False),
    0x24: ('home', False),
    0x25: ('left', False),
    0x26: ('up', False),
    0x27: ('right', False),
    0x28: ('down', False),
    0x29: ('select', False),
    0x2a: ('print', False),
    0x2b: ('execute', False),
    0x2c: ('print screen', False),
    0x2d: ('insert', False),
    0x2e: ('delete', False),
    0x2f: ('help', False),
    0x30: ('0', False),
    0x31: ('1', False),
    0x32: ('2', False),
    0x33: ('3', False),
    0x34: ('4', False),
    0x35: ('5', False),
    0x36: ('6', False),
    0x37: ('7', False),
    0x38: ('8', False),
    0x39: ('9', False),
    0x41: ('a', False),
    0x42: ('b', False),
    0x43: ('c', False),
    0x44: ('d', False),
    0x45: ('e', False),
    0x46: ('f', False),
    0x47: ('g', False),
    0x48: ('h', False),
    0x49: ('i', False),
    0x4a: ('j', False),
    0x4b: ('k', False),
    0x4c: ('l', False),
    0x4d: ('m', False),
    0x4e: ('n', False),
    0x4f: ('o', False),
    0x50: ('p', False),
    0x51: ('q', False),
    0x52: ('r', False),
    0x53: ('s', False),
    0x54: ('t', False),
    0x55: ('u', False),
    0x56: ('v', False),
    0x57: ('w', False),
    0x58: ('x', False),
    0x59: ('y', False),
    0x5a: ('z', False),
    0x5b: ('left windows', False),
    0x5c: ('right windows', False),
    0x5d: ('applications', False),
    0x5f: ('sleep', False),
    0x60: ('0', True),
    0x61: ('1', True),
    0x62: ('2', True),
    0x63: ('3', True),
    0x64: ('4', True),
    0x65: ('5', True),
    0x66: ('6', True),
    0x67: ('7', True),
    0x68: ('8', True),
    0x69: ('9', True),
    0x6a: ('*', True),
    0x6b: ('+', True),
    0x6c: ('separator', True),
    0x6d: ('-', True),
    0x6e: ('decimal', True),
    0x6f: ('/', True),
    0x70: ('f1', False),
    0x71: ('f2', False),
    0x72: ('f3', False),
    0x73: ('f4', False),
    0x74: ('f5', False),
    0x75: ('f6', False),
    0x76: ('f7', False),
    0x77: ('f8', False),
    0x78: ('f9', False),
    0x79: ('f10', False),
    0x7a: ('f11', False),
    0x7b: ('f12', False),
    0x7c: ('f13', False),
    0x7d: ('f14', False),
    0x7e: ('f15', False),
    0x7f: ('f16', False),
    0x80: ('f17', False),
    0x81: ('f18', False),
    0x82: ('f19', False),
    0x83: ('f20', False),
    0x84: ('f21', False),
    0x85: ('f22', False),
    0x86: ('f23', False),
    0x87: ('f24', False),
    0x90: ('num lock', True),
    0x91: ('scroll lock', False),
    0xa0: ('left shift', False),
    0xa1: ('right shift', False),
    0xa2: ('left ctrl', False),
    0xa3: ('right ctrl', False),
    0xa4: ('left menu', False),
    0xa5: ('right menu', False),
    0xa6: ('browser back', False),
    0xa7: ('browser forward', False),
    0xa8: ('browser refresh', False),
    0xa9: ('browser stop', False),
    0xaa: ('browser search key ', False),
    0xab: ('browser favorites', False),
    0xac: ('browser start and home', False),
    0xad: ('volume mute', False),
    0xae: ('volume down', False),
    0xaf: ('volume up', False),
    0xb0: ('next track', False),
    0xb1: ('previous track', False),
    0xb2: ('stop media', False),
    0xb3: ('play/pause media', False),
    0xb4: ('start mail', False),
    0xb5: ('select media', False),
    0xb6: ('start application 1', False),
    0xb7: ('start application 2', False),
    0xbb: ('+', False),
    0xbc: (',', False),
    0xbd: ('-', False),
    0xbe: ('.', False),
    # 0xbe: ('/', False), # Used for miscellaneous characters; it can vary by keyboard. For the US standard keyboard, the '/?' key.
    0xe5: ('ime process', False),
    0xf6: ('attn', False),
    0xf7: ('crsel', False),
    0xf8: ('exsel', False),
    0xf9: ('erase eof', False),
    0xfa: ('play', False),
    0xfb: ('zoom', False),
    0xfc: ('reserved ', False),
    0xfd: ('pa1', False),
    0xfe: ('clear', False),
}

canonical_names = {
    'escape': 'esc',
    'return': 'enter',
    'del': 'delete',
    'control': 'ctrl',
    'altgr': 'alt gr',

    'left arrow': 'left',
    'up arrow': 'up',
    'down arrow': 'down',
    'right arrow': 'right',

    ' ': 'space',  # Prefer to spell out keys that would be hard to read.
    '\x1b': 'esc',
    '\x08': 'backspace',
    '\n': 'enter',
    '\t': 'tab',
    '\r': 'enter',

    'scrlk': 'scroll lock',
    'prtscn': 'print screen',
    'prnt scrn': 'print screen',
    'snapshot': 'print screen',
    'ins': 'insert',
    'pause break': 'pause',
    'ctrll lock': 'caps lock',
    'capslock': 'caps lock',
    'number lock': 'num lock',
    'numlock:': 'num lock',
    'space bar': 'space',
    'spacebar': 'space',
    'linefeed': 'enter',
    'win': 'windows',

    'app': 'menu',
    'apps': 'menu',
    'application': 'menu',
    'applications': 'menu',

    'pagedown': 'page down',
    'pageup': 'page up',
    'pgdown': 'page down',
    'pgup': 'page up',
    'next': 'page down',  # This looks wrong, but this is how Linux reports.
    'prior': 'page up',

    'underscore': '_',
    'equal': '=',
    'minplus': '+',
    'plus': '+',
    'add': '+',
    'subtract': '-',
    'minus': '-',
    'multiply': '*',
    'asterisk': '*',
    'divide': '/',

    'question': '?',
    'exclam': '!',
    'slash': '/',
    'bar': '|',
    'backslash': '\\',
    'braceleft': '{',
    'braceright': '}',
    'bracketleft': '[',
    'bracketright': ']',
    'parenleft': '(',
    'parenright': ')',

    'period': '.',
    'dot': '.',
    'comma': ',',
    'semicolon': ';',
    'colon': ':',

    'less': '<',
    'greater': '>',
    'ampersand': '&',
    'at': '@',
    'numbersign': '#',
    'hash': '#',
    'hashtag': '#',

    'dollar': '$',
    'sterling': '£',
    'pound': '£',
    'yen': '¥',
    'euro': '€',
    'cent': '¢',
    'currency': '¤',
    'registered': '®',
    'copyright': '©',
    'notsign': '¬',
    'percent': '%',
    'diaeresis': '"',
    'quotedbl': '"',
    'onesuperior': '¹',
    'twosuperior': '²',
    'threesuperior': '³',
    'onehalf': '½',
    'onequarter': '¼',
    'threequarters': '¾',
    'paragraph': '¶',
    'section': '§',
    'ssharp': '§',
    'division': '÷',
    'questiondown': '¿',
    'exclamdown': '¡',
    'degree': '°',
    'guillemotright': '»',
    'guillemotleft': '«',

    'acute': '´',
    'agudo': '´',
    'grave': '`',
    'tilde': '~',
    'asciitilde': '~',
    'til': '~',
    'cedilla': ',',
    'circumflex': '^',
    'apostrophe': '\'',

    'adiaeresis': 'ä',
    'udiaeresis': 'ü',
    'odiaeresis': 'ö',
    'oe': 'Œ',
    'oslash': 'ø',
    'ooblique': 'Ø',
    'ccedilla': 'ç',
    'ntilde': 'ñ',
    'eacute': 'é',
    'uacute': 'ú',
    'oacute': 'ó',
    'thorn': 'þ',
    'ae': 'æ',
    'eth': 'ð',
    'masculine': 'º',
    'feminine': 'ª',
    'iacute': 'í',
    'aacute': 'á',
    'mu': 'Μ',
    'aring': 'å',

    'zero': '0',
    'one': '1',
    'two': '2',
    'three': '3',
    'four': '4',
    'five': '5',
    'six': '6',
    'seven': '7',
    'eight': '8',
    'nine': '9',

    'play/pause': 'play/pause media',

    'num multiply': '*',
    'num divide': '/',
    'num add': '+',
    'num plus': '+',
    'num minus': '-',
    'num sub': '-',
    'num enter': 'enter',
    'num 0': '0',
    'num 1': '1',
    'num 2': '2',
    'num 3': '3',
    'num 4': '4',
    'num 5': '5',
    'num 6': '6',
    'num 7': '7',
    'num 8': '8',
    'num 9': '9',

    'left win': 'left windows',
    'right win': 'right windows',
    'left control': 'left ctrl',
    'right control': 'right ctrl',
    'left menu': 'left alt',  # Windows...
}


def normalize_name(name):
    if not name:
        return 'unknown'
    name = name.lower()
    if name != '_':
        name = name.replace('_', ' ')
    return canonical_names.get(name, name)


def _create_key_name_tables():
    """
    Build tables for scan codes.

    Returns:
        lower_case: `dict` mapping scan_code to lower-case key names
        upper_case: `dict` mapping scan code to upper-case key names
    """
    key_names_by_scan_code = {}
    scan_code_to_vk = {}

    for vk in range(0x01, 0x100):
        scan_code = MapVirtualKey(vk, MAPVK_VK_TO_VSC)
        if not scan_code: continue

        # Scan codes may map to multiple virtual key codes.
        # In this case prefer the officially defined ones.
        if scan_code_to_vk.get(scan_code, 0) not in VIRTUAL_KEYBOARD:
            scan_code_to_vk[scan_code] = vk

    name_buffer = ctypes.create_unicode_buffer(32)
    keyboard_state = keyboard_state_type()
    for scan_code in range(2 ** (23 - 16)):
        key_names_by_scan_code[scan_code] = ['unknown', 'unknown']

        # Get pure key name, such as "shift". This depends on locale and
        # may return a translated name.
        for enhanced in [1, 0]:
            ret = GetKeyNameText(scan_code << 16 | enhanced << 24, name_buffer, 1024)
            if not ret:
                continue
            name = normalize_name(name_buffer.value)
            key_names_by_scan_code[scan_code] = [name, name]

        if scan_code not in scan_code_to_vk: continue
        # Get associated character, such as "^", possibly overwriting the pure key name.
        for shift_state in [0, 1]:
            keyboard_state[0x10] = shift_state * 0xFF
            vk = scan_code_to_vk.get(scan_code, 0)
            ret = ToUnicode(vk, scan_code, keyboard_state, name_buffer, len(name_buffer), 0)
            if ret:
                # Sometimes two characters are written before the char we want,
                # usually an accented one such as Â. Couldn't figure out why.
                char = name_buffer.value[-1]
                key_names_by_scan_code[scan_code][shift_state] = char

    key_names_by_scan_code[541] = ['alt gr', 'alt gr']
    # key_names_by_vk_code = {vk: key_names_by_scan_code[scan_code] for scan_code, vk in scan_code_to_vk.items()}
    return {c: lower for c, (lower, upper) in key_names_by_scan_code.items()}, {c: upper for c, (lower, upper) in key_names_by_scan_code.items()}


KEY_NAMES_LOWER, KEY_NAMES_UPPER = _create_key_name_tables()

_REGISTERED_PROCEDURES_REFS = []  # do not garbage collect these


def set_window_procedure(hwnd, procedure: Callable, call_original=False):
    def process_message(hwnd, msg, wParam, lParam):
        procedure(hwnd, msg, wParam, lParam)
        if call_original:
            return CallWindowProc(prevWndProc, hwnd, msg, wParam, lParam)
        else:
            return user32.DefWindowProcA(c_int(hwnd), c_int(msg), c_int(wParam), c_int(lParam))

    GWL_WNDPROC = ctypes.c_int(-4)
    new_window_procedure = WNDPROCTYPE(process_message)
    _REGISTERED_PROCEDURES_REFS.append(new_window_procedure)
    prevWndProc = user32.SetWindowLongPtrA(hwnd, GWL_WNDPROC, new_window_procedure)
    if not prevWndProc:
        raise ctypes.WinError(GetLastError())


def invisible_window(procedure: Callable):
    def actual_procedure(hwnd, msg, wParam, lParam):
        procedure(hwnd, msg, wParam, lParam)
        return user32.DefWindowProcA(c_int(hwnd), c_int(msg), c_int(wParam), c_int(lParam))

    h_instance = kernel32.GetModuleHandleW(0)
    window_class = WNDCLASSEX()
    window_class.cbSize = sizeof(WNDCLASSEX)
    window_class.style = CS_HREDRAW | CS_VREDRAW
    window_class.lpfnWndProc = WNDPROCTYPE(actual_procedure)
    window_class.cbClsExtra = 0
    window_class.cbWndExtra = 0
    window_class.hInstance = h_instance
    window_class.hIcon = 0
    window_class.hCursor = 0
    window_class.hBrush = ctypes.windll.gdi32.GetStockObject(WHITE_BRUSH)
    window_class.lpszMenuName = 0
    window_class.lpszClassName = 'Python Win32 Class'
    window_class.hIconSm = 0
    if not user32.RegisterClassExW(byref(window_class)):
        raise ctypes.WinError(GetLastError())
    h_window = CreateWindowExW(
        0,  # dwExStyle
        window_class.lpszClassName,
        'Invisible Window for Python Win32',
        0,  # dwStyle
        0, 0, 0, 0,  # x, y, nWidth, nHeight,
        0, 0, h_instance, 0)
    if not h_window:
        raise RuntimeError(f"Failed to create window: error code {GetLastError()}")
    _REGISTERED_PROCEDURES_REFS.append(window_class)
    return h_window


def enable_raw_input_for_window(hwnd: int, device_type='Keyboard'):
    usage_page, usage = USAGE_NAME_TO_VALUE[device_type.lower()]
    raw_input_device = RAWINPUTDEVICE()
    raw_input_device.us_usage_page = usage_page
    raw_input_device.us_usage = usage
    raw_input_device.dw_flags = RIDEV_INPUTSINK
    raw_input_device.hwnd_target = hwnd  # 0 to follow the keyboard focus, else hWnd
    if not RegisterRawInputDevices(raw_input_device, 1, sizeof(RAWINPUTDEVICE)):
        raise ctypes.WinError(GetLastError())


MOUSE_BUTTONS = {
    1: ('down', 1, 'left'),
    2: ('up', 1, 'left'),
    4: ('down', 3, 'right'),
    8: ('up', 3, 'right'),
    16: ('down', 2, 'middle'),  # wheel pressed
    32: ('up', 2, 'middle'),  # wheel released
    64: ('down', 4, 'thumb1'),
    128: ('up', 4, 'thumb1'),
    256: ('down', 5, 'thumb2'),
    512: ('up', 5, 'thumb2'),
    7865344: ('wheel-up', 2, 'wheel'),  # wheel turned
    4287104000: ('wheel-down', 2, 'wheel')  # wheel turned
}

KEY_EVENT_TYPE = {256: 'down', 257: 'up', 260: 'down'}  # 260 for multi-key-down (alt gr)

MOUSE_TYPES = {
    0x0080: 'HID mouse',
    0x0100: 'HID wheel mouse',
    0x8000: 'Mouse with horizontal wheel',
}

MOVE_MODES = {
    0x00: 'relative',
    0x01: 'absolute',
    0x02: 'virtual desktop',
    0x04: 'attributes changed',
    0x08: 'no coalesce',
}

KEYBOARD_TYPES = {
    0x4: "Enhanced 101- or 102-key keyboards (and compatibles)",
    0x7: "Japanese Keyboard",
    0x8: "Korean Keyboard",
    0x51: "Unknown type or HID keyboard",
}

GENERIC_DESKTOP_CONTROLS_PAGE = {
    0x01:	"Pointer",
    0x02:	"Mouse",
    0x04:	"Joystick",
    0x05:	"Game Pad",
    0x06:	"Keyboard",
    0x07:	"Keypad",
    0x08:	"Multi-axis Controller"
}
USAGE_NAME_TO_VALUE = {v.lower(): (1, k) for k, v in GENERIC_DESKTOP_CONTROLS_PAGE.items()}

USAGE_PAGE_NAMES = {
    0x01: ("Generic Desktop Controls", GENERIC_DESKTOP_CONTROLS_PAGE),
    0x05: ("Game Controls", {}),
    0x08: ("LEDs", {}),
    0x09: ("Button", {}),
}


def get_raw_input_device_list() -> ctypes.Array:
    device_count = ctypes.c_uint()
    if GetRawInputDeviceList(None, device_count, ctypes.sizeof(RAWINPUTDEVICELIST)) == -1:
        raise ctypes.WinError()
    devices = (RAWINPUTDEVICELIST * device_count.value)()
    if GetRawInputDeviceList(devices, device_count, ctypes.sizeof(RAWINPUTDEVICELIST)) == -1:
        raise ctypes.WinError()
    return devices


def get_device_path(device) -> Optional[str]:
    """
    Args:
        device: RAWINPUTDEVICELIST or hDevice

    Returns:
        `RIDI_DEVICENAME` as `str`. This string typically includes vendor and product id.
    """
    if device is None:
        return None
    device_handle = device.hDevice if isinstance(device, RAWINPUTDEVICELIST) else device
    name_size = ctypes.c_uint()  # Get the size of the device name buffer
    if GetRawInputDeviceInfoW(device_handle, RIDI_DEVICENAME, None, name_size):
        raise ctypes.WinError(GetLastError())
    name_buffer = (WCHAR * name_size.value)()
    if not GetRawInputDeviceInfoW(device_handle, RIDI_DEVICENAME, name_buffer, name_size):
        raise ctypes.WinError(GetLastError())
    return name_buffer.value


def is_connected(device) -> bool:
    """
    Args:
        device: RAWINPUTDEVICELIST or hDevice

    Returns:
        Whether the device is currently available.
    """
    device_handle = device.hDevice if isinstance(device, RAWINPUTDEVICELIST) else device
    name_size = ctypes.c_uint()  # Get the size of the device name buffer
    ret = GetRawInputDeviceInfoW(device_handle, RIDI_DEVICENAME, None, name_size)
    if ret == 0:
        return True
    elif GetLastError() == 6:
        return False
    else:
        raise ctypes.WinError(GetLastError())


def get_device_info(device):
    """
    Args:
        device_handle: RAWINPUTDEVICELIST or hDevice

    Returns:
        `RIDI_DEVICEINFO` struct
    """
    if device is None:
        return RID_DEVICE_INFO()
    device_handle = device.hDevice if isinstance(device, RAWINPUTDEVICELIST) else device
    info_size = ctypes.c_uint()
    if GetRawInputDeviceInfoW(device_handle, RIDI_DEVICEINFO, None, info_size):
        raise ctypes.WinError()
    info_buffer = (RID_DEVICE_INFO * info_size.value)()
    if not GetRawInputDeviceInfoW(device_handle, RIDI_DEVICEINFO, info_buffer, info_size):
        raise ctypes.WinError()
    return info_buffer[0]



from setuptools import setup

long_description = """
Enumerate raw input devices and receive input events with device ID on Windows.

Homepage: https://github.com/holl-/win-raw-in
"""

setup(
    name='win-raw-in',
    version='0.2.0',
    author='Philipp Holl',
    author_email='philipp@mholl.de',
    packages=['winrawin'],
    package_data={'': ['winrawin/usb.ids']},
    include_package_data=True,
    url='https://github.com/holl-/win-raw-in',
    license='MIT',
    description='Enumerate raw input devices and receive input events with device ID on Windows',
    keywords='keyboard mouse hook raw input',
    long_description=long_description,
    install_requires=['dataclasses'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ],
)



import os.path
import re
from typing import Tuple, Optional


with open(os.path.join(os.path.dirname(__file__), 'usb.ids'), 'r') as usb_ids_file:
    DATABASE = usb_ids_file.read()


def lookup_product(vendor_id: int, product_id: int) -> Tuple[Optional[str], Optional[str]]:
    vendor_id = f'{vendor_id:#06x}'[2:]
    product_id = f'{product_id:#06x}'[2:]
    try:
        vendor_pos = DATABASE.index(f'\n{vendor_id}')
    except ValueError:
        return None, None
    vendor_pos_end = DATABASE.index('\n', vendor_pos+7)
    vendor_name = DATABASE[vendor_pos+7:vendor_pos_end]
    next_vendor_pos = re.search(r'\n[0-9a-zA-Z]', DATABASE[vendor_pos_end:]).start()  # ToDo if this is the last vendor, we get an error
    products_database = DATABASE[vendor_pos_end:vendor_pos_end+next_vendor_pos]
    try:
        product_pos = products_database.index(f'\n\t{product_id}')
        product_pos_end = products_database.index('\n', product_pos+7)
        product_name = products_database[product_pos+7:product_pos_end]
    except ValueError:
        product_name = None
    return vendor_name, product_name
    # ToDo if not in database, download new database http://www.linux-usb.org/usb.ids


def parse_vid_pid(device_id: str):
    if device_id is None:
        return 0, 0
    vendor_id = re.search(r'VID_(.*?)[&#]', device_id).group(1)
    product_id = re.search(r'PID_(.*?)[&#]', device_id).group(1)
    return int(vendor_id, 16), int(product_id, 16)


if __name__ == '__main__':
    vid, pid = parse_vid_pid('\\\\?\\HID#{00001812-0000-1000-8000-00805f9b34fb}&Dev&VID_045e&PID_0b13&REV_0509&0c352633ee04&IG_00#a&3724ae32&0&0000#{4d1e55b2-f16f-11cf-88cb-001111000030}')
    print(lookup_product(vendor_id=vid, product_id=pid))
